@@ -1,5 +1,5 @@
 /* Code to test for "definitive [un]assignment".
-   Copyright (C) 1999, 2000, 2001, 2003, 2004, 2005 Free Software Foundation,
+   Copyright (C) 1999, 2000, 2001, 2003, 2004, 2005, 2006 Free Software Foundation,
    Inc.
 
 This file is part of GCC.
@@ -672,8 +672,7 @@ check_init (tree exp, words before)
 	DECLARE_BUFFERS(when_true, 2);
 	words when_false = when_true + num_current_words;
 #ifdef ENABLE_JC1_CHECKING
-	if (TREE_CODE (alt->block) != LOOP_EXPR)
-	  abort ();
+	gcc_assert (TREE_CODE (alt->block) == LOOP_EXPR);
 #endif
 	check_bool_init (TREE_OPERAND (exp, 0), before, when_false, when_true);
 	done_alternative (when_true, alt);
@@ -807,6 +806,7 @@ check_init (tree exp, words before)
     case TRUTH_NOT_EXPR:
     case BIT_NOT_EXPR:
     case CONVERT_EXPR:
+    case VIEW_CONVERT_EXPR:
     case BIT_FIELD_REF:
     case FLOAT_EXPR:
     case FIX_TRUNC_EXPR:
